#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <xf86drm.h>
#include "nouveau_drm.h"
#include "nouveau_device.h"
#include "nouveau_notifier.h"
#include "nouveau_channel.h"
#include "nouveau_pushbuf.h"
#include "nouveau_bo.h"
#include "nouveau_class.h"

#define NvDmaFB 0xD0FB0001
#define NvDmaTT 0xD0AA0001
#define Nv2D 0xD02D0001
#define NvNotify 0xD0000001

struct nouveau_device *dev = NULL;
struct nouveau_channel *chan = NULL;
struct nouveau_notifier *notifier = NULL;
struct nouveau_grobj *eng2d = NULL;
struct nouveau_bo *bo = NULL;

#define WIDTH 50
#define HEIGHT 100
#define TILING_PARAM 0x40
#define TILE_WIDTH 64
#define TILE_HEIGHT 64
#define PITCH_WIDTH ((WIDTH + 63) & ~63)
#define PITCH_HEIGHT ((HEIGHT + TILE_HEIGHT - 1) & ~(TILE_HEIGHT - 1))

void
get_dst()
{
	int ret = 0, i;
	uint8_t *map;

	if (ret = nouveau_bo_new(dev, NOUVEAU_BO_VRAM | NOUVEAU_BO_TILED, 256, PITCH_WIDTH*PITCH_HEIGHT, &bo)) {
		fprintf(stderr, "nouveau_bo_new failed with %d\n", ret);
		return;
	}

	/* Set destination. */
	BEGIN_RING (chan, eng2d, NV50_2D_DST_FORMAT, 5);
	OUT_RING (chan, NV50_2D_SRC_FORMAT_8BPP);
	OUT_RING (chan, 0);
	OUT_RING (chan, TILING_PARAM);
	OUT_RING  (chan, 1);
	OUT_RING  (chan, 0);

	BEGIN_RING (chan, eng2d, NV50_2D_DST_WIDTH, 4);
	OUT_RING (chan, WIDTH);
	OUT_RING (chan, HEIGHT);
	OUT_RELOCh(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, NV50_2D_CLIP_X, 4);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, WIDTH);
	OUT_RING  (chan, HEIGHT);

	BEGIN_RING(chan, eng2d, NV50_2D_OPERATION, 1);
	OUT_RING  (chan, NV50_2D_OPERATION_SRCCOPY);

	FIRE_RING  (chan);

	/* clear the pixmap */
	nouveau_bo_map(bo, NOUVEAU_BO_RDWR);
	map = bo->map;
	if (!map) {
		fprintf(stderr, "map failed\n");
		return;
	}

	for (i = 0; i < bo->size; i++)
		map[i] = 0x00;

	nouveau_bo_unmap(bo);
}

void
set_pixel(int x, int y, uint32_t pixel)
{
	BEGIN_RING (chan, eng2d, NV50_2D_DST_ADDRESS_HIGH, 2);
	OUT_RELOCh(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, 0x580, 3);
	OUT_RING  (chan, 4);
	OUT_RING  (chan, NV50_2D_SRC_FORMAT_8BPP);
	OUT_RING  (chan, pixel);

	BEGIN_RING(chan, eng2d, NV50_2D_RECT_X1, 4);
	OUT_RING  (chan, x);
	OUT_RING  (chan, y);
	OUT_RING  (chan, x + 1);
	OUT_RING  (chan, y + 1);

	FIRE_RING (chan);
}

void
find_pixel (int *offset, uint32_t pixel)
{
	uint8_t *map = NULL;

	nouveau_bo_map(bo, NOUVEAU_BO_RDWR);
	map = bo->map;
	if (!map) {
		fprintf(stderr, "map failed\n");
		return;
	}

	for ((*offset) = 0; (*offset) < PITCH_WIDTH*PITCH_HEIGHT; (*offset)++)
		if ((map[(*offset)]) == ((pixel) & 0xFF))
			goto out;

out:
	map[(*offset)] = 0x00;
	nouveau_bo_unmap(bo);
}

int
recurse_offset(int offset, int tile_height, int x, int y, int old_tile_width)
{
	int tile_width = 0;
	if (tile_height == 2)
		tile_width = 8;
	else if (tile_height == 4)
		tile_width = 32;
	else
		tile_width = 64;

	offset += ((x/tile_width) + (y/tile_height) * (old_tile_width/tile_width)) * (tile_height *  tile_width);

	if (tile_height > 2)
		offset = recurse_offset(offset, tile_height/2, x % tile_width, y % tile_height, tile_width);
	else
		offset += (y % tile_height) * tile_width + (x % tile_width);

	return offset;
}

int
main(int argc, char **argv)
{
	int ret = 0;
	int x = 0, y = 0;
	int offset = 0, offset_expected = 0;

	if (ret = nouveau_device_open(&dev, NULL)) {
		fprintf(stderr, "nouveau_device_open failed\n");
		return ret;
	}

	if (ret = nouveau_channel_alloc(dev, NvDmaFB, NvDmaTT, &chan)) {
		fprintf(stderr, "nouveau_channel_alloc failed\n");
		return ret;
	}

	if (ret = nouveau_notifier_alloc(chan, NvNotify, 1, &notifier)) {
		fprintf(stderr, "nouveau_notifier_alloc failed\n");
		return ret;
	}

	/* 2D object */
	if (ret = nouveau_grobj_alloc(chan, Nv2D, 0x502d, &eng2d)) {
		fprintf(stderr, "nouveau_grobj_alloc failed\n");
		return ret;
	}

	BEGIN_RING (chan, eng2d, 0x180, 3);
	OUT_RING  (chan, notifier->handle);
	OUT_RING  (chan, chan->vram->handle);
	OUT_RING  (chan, chan->vram->handle);

	/* Magics from nv, no clue what they do, but at least some
	 * of them are needed to avoid crashes.
	 */
	BEGIN_RING(chan, eng2d, 0x260, 1);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, 0x290, 1);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, 0x29c, 1);
	OUT_RING  (chan, 0);
	BEGIN_RING(chan, eng2d, 0x58c, 1);
	OUT_RING  (chan, 0x111);

	/* Now we are ready for the real stuff. */
	get_dst();
	if (!bo) {
		fprintf(stderr, "get_dst failed\n");
		return -ENOMEM;
	}

	for (y = 0; y < HEIGHT; y++) {
		for (x = 0; x < WIDTH; x++) {
			set_pixel(x, y, 0xFFFFFFFF);
			find_pixel(&offset, 0xFFFFFFFF);
			offset_expected = recurse_offset(0, TILE_HEIGHT, x, y, PITCH_WIDTH);
			if (offset_expected != offset) {
				printf("offset %d offset_expected %d\n", offset, offset_expected);
				printf("x %d y %d\n", x, y);
			}
			//set_pixel(x, y, 0x00000000);
		}
	}

	nouveau_device_close(&dev);

	return ret;
}
